from typing import Union

from py_near.dapps.ft.exceptions import NotRegisteredError, NotEnoughBalance
from py_near.exceptions.exceptions import FunctionCallError

from py_near.dapps.core import DappClient, NEAR
from py_near.dapps.ft.models import FtTokenMetadata
from py_near.dapps.fts import FtModel
from typing import Optional


class FT(DappClient):
    async def get_ft_balance(
        self, ft: FtModel, account_id: Optional[str] = None
    ) -> float:
        """
        Get fungible token balance
        :param ft: fungible token model FT.USDC
        :param account_id: account id
        :return: amount // 10**ft.decimal
        """
        if not account_id:
            account_id = self._account.account_id
        return (
            await self.get_ft_raw_balance(ft.contract_id, account_id) / 10**ft.decimal
        )

    async def get_ft_raw_balance(
        self, contract_id: str, account_id: Optional[str] = None
    ) -> int:
        """
        Get fungible token raw balance
        :param contract_id: fungible token contract adress
        :param account_id: account id
        :return: amount
        """
        if not account_id:
            account_id = self._account.account_id
        return int(
            (
                await self._account.view_function(
                    contract_id,
                    "ft_balance_of",
                    {"account_id": account_id},
                )
            ).result
            or 0
        )

    async def get_metadata(self, ft: Union[FtModel, str]) -> FtTokenMetadata:
        """
        Get fungible token metadata
        :param ft: fungible token model FT.USDC
        :return: FtTokenMetadata
        """
        if isinstance(ft, FtModel):
            contract_id = ft.contract_id
        else:
            contract_id = ft
        return FtTokenMetadata(
            **(
                await self._account.view_function(
                    contract_id,
                    "ft_metadata",
                    {},
                )
            ).result
        )

    async def transfer(
        self,
        ft: FtModel,
        receiver_id: str,
        amount: float,
        memo: str = "",
        force_register: bool = False,
        nowait: bool = False,
    ):
        """
        Transfer fungible token to account

        :param ft: fungible token model FT.USDC
        :param receiver_id: receiver account id
        :param amount: float amount to transfer. 1 for 1 USDC
        :param memo: comment
        :param force_register: use storage_deposit() if account is not registered
        :return: transaction hash ot TransactionResult
        """
        if (
            force_register
            and await self.storage_balance_of(ft, receiver_id) < NEAR // 500
        ):
            await self.storage_deposit(ft, receiver_id)
        try:
            return await self._account.function_call(
                ft.contract_id,
                "ft_transfer",
                {
                    "receiver_id": receiver_id,
                    "amount": str(int(amount * 10**ft.decimal)),
                    "msg": memo,
                },
                amount=1,
                nowait=nowait,
            )
        except FunctionCallError as e:
            if "The account is not registered" in e.error["ExecutionError"]:
                raise NotRegisteredError(
                    "The receiver is not registered, user .storage_deposit() method to register it"
                )
            if "The account doesn't have enough balance" in e.error["ExecutionError"]:
                raise NotEnoughBalance(e)
            raise e

    async def transfer_call(
        self,
        ft: FtModel,
        receiver_id: str,
        amount: float,
        memo: str = "",
        force_register: bool = False,
        nowait: bool = False,
    ):
        """
        Transfer fungible token to account and call ft_on_transfer() method in receiver contract

        :param ft: fungible token model FT.USDC
        :param receiver_id: receiver account id
        :param amount: float amount to transfer. 1 for 1 USDC
        :param memo: comment
        :param force_register: use storage_deposit() if account is not registered
        :param nowait if True, method will return before transaction is confirmed
        :return: transaction hash ot TransactionResult
        """
        if (
            force_register
            and await self.storage_balance_of(ft, receiver_id) < NEAR // 500
        ):
            await self.storage_deposit(ft, receiver_id)
        return await self._account.function_call(
            ft.contract_id,
            "ft_transfer_call",
            {
                "receiver_id": receiver_id,
                "amount": str(int(amount * 10**ft.decimal)),
                "msg": memo,
            },
            amount=1,
            nowait=nowait,
        )

    async def storage_balance_of(
        self, ft: Union[FtModel, str], account_id: Optional[str] = None
    ) -> int:
        """
        Get storage balance of account. The balance must be greater than 0.01 NEAR for some smart contracts
        in order for the recipient to accept the token

        :param contract_id: fungible token contract_id
        :param account_id: account id
        :return: int balance in yoctoNEAR, 1_000_000_000_000_000_000_000_000 for 1 NEAR
        """
        if not account_id:
            account_id = self._account.account_id
        if isinstance(ft, FtModel):
            contract_id = ft.contract_id
        else:
            contract_id = ft
        res = (
            await self._account.view_function(
                contract_id,
                "storage_balance_of",
                {"account_id": account_id},
            )
        ).result
        if res:
            return int(res["total"] or 0)
        return 0

    async def storage_deposit(
        self,
        ft: Union[FtModel, str],
        account_id: Optional[str] = None,
        amount: int = NEAR // 50,
    ):
        """
        Deposit storage balance for account. The balance must be greater than 0.01 NEAR for some smart contracts

        :param ft: fungible token model FT.USDC
        :param account_id: receiver account id
        :param amount: in amount of yoctoNEAR
        :return:
        """
        if not account_id:
            account_id = self._account.account_id
        if isinstance(ft, FtModel):
            contract_id = ft.contract_id
        else:
            contract_id = ft
        return await self._account.function_call(
            contract_id,
            "storage_deposit",
            {"account_id": account_id},
            amount=amount,
        )



import asyncio
import base64
import datetime
import json
from collections import Counter
from typing import Optional

import aiohttp
from aiohttp import ClientResponseError, ClientConnectorError, ServerDisconnectedError
from loguru import logger

from py_near.constants import TIMEOUT_WAIT_RPC
from py_near.exceptions.exceptions import RpcNotAvailableError, RpcEmptyResponse
from py_near.exceptions.provider import (
    UnknownBlockError,
    InvalidAccount,
    NoContractCodeError,
    UnknownAccount,
    TooLargeContractStateError,
    UnavailableShardError,
    NoSyncedBlocksError,
    InternalError,
    NoSyncedYetError,
    InvalidTransactionError,
    RPCTimeoutError,
    UnknownAccessKeyError,
    ERROR_CODE_TO_EXCEPTION,
    InvalidNonce,
)
from py_near.models import TransactionResult

PROVIDER_CODE_TO_EXCEPTION = {
    "UNKNOWN_BLOCK": UnknownBlockError,
    "INVALID_ACCOUNT": InvalidAccount,
    "UNKNOWN_ACCOUNT": UnknownAccount,
    "NO_CONTRACT_CODE": NoContractCodeError,
    "TOO_LARGE_CONTRACT_STATE": TooLargeContractStateError,
    "UNAVAILABLE_SHARD": UnavailableShardError,
    "NO_SYNCED_BLOCKS": NoSyncedBlocksError,
    "INTERNAL_ERROR": InternalError,
    "NOT_SYNCED_YET": NoSyncedYetError,
    "INVALID_TRANSACTION": InvalidTransactionError,
    "TIMEOUT_ERROR": RPCTimeoutError,
    "UNKNOWN_ACCESS_KEY": UnknownAccessKeyError,
}


class JsonProvider(object):
    def __init__(self, rpc_addr, allow_broadcast=True, timeout=TIMEOUT_WAIT_RPC):
        """
        :param rpc_addr: str or list of str
        :param allow_broadcast: bool - submit signed transaction to all RPCs
        """
        if isinstance(rpc_addr, tuple):
            self._rpc_addresses = ["http://{}:{}".format(*rpc_addr)]
        elif isinstance(rpc_addr, list):
            self._rpc_addresses = rpc_addr
        else:
            self._rpc_addresses = [rpc_addr]
        self._available_rpcs = self._rpc_addresses.copy()
        self._last_rpc_addr_check = 0
        self.allow_broadcast = allow_broadcast
        self._timeout = timeout
        self.session = aiohttp.ClientSession()

    async def shutdown(self):
        await self.session.close()

    async def check_available_rpcs(self):
        if (
            self._last_rpc_addr_check < datetime.datetime.now().timestamp() - 30
            or not self._available_rpcs
        ):
            self._last_rpc_addr_check = datetime.datetime.now().timestamp()
            asyncio.create_task(self._check_available_rpcs())

        for _ in range(5):
            if self._available_rpcs:
                break
            await self._check_available_rpcs()
            await asyncio.sleep(3)

        if not self._available_rpcs:
            raise RpcNotAvailableError("All RPCs are unavailable")

    async def _check_available_rpcs(self):
        available_rpcs = []
        for rpc_addr in self._rpc_addresses:
            try:
                data = {
                    "jsonrpc": "2.0",
                    "method": "status",
                    "params": {"finality": "final"},
                    "id": 1,
                }
                async with self.session.post(rpc_addr, json=data) as r:
                    if r.status == 200:
                        data = json.loads(await r.text())["result"]
                        if data["sync_info"]["syncing"]:
                            last_block_ts = datetime.datetime.fromisoformat(
                                data["sync_info"]["latest_block_time"]
                            )
                            diff = (
                                datetime.datetime.utcnow().timestamp()
                                - last_block_ts.timestamp()
                            )
                            is_syncing = diff > 60
                        else:
                            is_syncing = False
                        if is_syncing:
                            logger.error(f"Remove async RPC : {rpc_addr} ({diff})")
                            continue
                        available_rpcs.append(rpc_addr)
                    else:
                        logger.error(
                            f"Remove rpc because of error {r.status}: {rpc_addr}"
                        )
            except Exception as e:
                if rpc_addr in self._available_rpcs:
                    logger.error(f"Remove rpc: {e}")
        self._available_rpcs = available_rpcs

    @staticmethod
    def most_frequent_by_hash(array):
        counter = Counter(array)
        most_frequent = counter.most_common(1)[0][0]
        return most_frequent

    async def call_rpc_request(
        self, method, params, broadcast=False, threshold: int = 0
    ):
        await self.check_available_rpcs()
        j = {"method": method, "params": params, "id": "dontcare", "jsonrpc": "2.0"}

        async def f(rpc_call_addr):
            auth_key = "py-near"
            if "@" in rpc_call_addr:
                auth_key = rpc_call_addr.split("//")[1].split("@")[0]
                rpc_call_addr = rpc_call_addr.replace(auth_key + "@", "")
            r = await self.session.post(
                rpc_call_addr,
                json=j,
                timeout=self._timeout,
                headers={
                    "Referer": "https://tgapp.herewallet.app",
                    "Authorization": f"Bearer {auth_key}",
                },  # NEAR RPC requires Referer header
            )
            if r.status == 200:
                return json.loads(await r.text())

        if broadcast or threshold:
            pending = [
                asyncio.create_task(f(rpc_addr)) for rpc_addr in self._available_rpcs
            ]

            responses = []
            while pending and len(pending):
                done, pending = await asyncio.wait(
                    pending, return_when=asyncio.FIRST_COMPLETED
                )
                for task in done:
                    try:
                        result = task.result()
                        if "error" not in result and (not threshold or threshold <= 1):
                            return result
                        responses.append(result)
                    except Exception as e:
                        logger.warning(e)
                if responses:
                    array = [hash(json.dumps(x)) for x in responses]
                    most_frequent_element = self.most_frequent_by_hash(array)
                    correct_responses = [
                        x for x in responses if hash(json.dumps(x)) == most_frequent_element
                    ]
                    if len(correct_responses) >= threshold:
                        for task in pending:
                            task.cancel()
                        return most_frequent_element
        else:
            for rpc_addr in self._available_rpcs:
                try:
                    res = await f(rpc_addr)
                    if "error" not in res:
                        return res
                except Exception as e:
                    logger.error(f"Rpc error: {e}")
                    continue
        raise RpcEmptyResponse("RPC returned empty response")

    @staticmethod
    def get_error_from_response(content: dict):
        if "error" in content:
            error_code = content["error"].get("cause", {}).get("name", "")
            body = content["error"]["data"]
            error = PROVIDER_CODE_TO_EXCEPTION.get(error_code, InternalError)(
                body, error_json=content["error"]
            )
            while True:
                if not isinstance(body, dict):
                    break
                if not body:
                    return error
                key, body = list(body.items())[0]
                if key in ERROR_CODE_TO_EXCEPTION:
                    error = ERROR_CODE_TO_EXCEPTION[key](
                        body, error_json=content["error"]
                    )
                else:
                    break
            return error

    async def json_rpc(self, method, params, broadcast=False, threshold=None):
        content = await self.call_rpc_request(
            method, params, broadcast=broadcast, threshold=threshold
        )
        if not content:
            raise RpcEmptyResponse("RPC returned empty response")

        error = self.get_error_from_response(content)
        if error:
            raise error
        return content["result"]

    async def send_tx(self, signed_tx: str):
        """
        Send a signed transaction to the network and return the hash of the transaction
        :param signed_tx: base64 encoded signed transaction, str.
        :param timeout: rpc request timeout
        :return:
        """
        return await self.json_rpc(
            "broadcast_tx_async",
            [signed_tx],
            broadcast=self.allow_broadcast,
        )

    async def send_tx_included(self, signed_tx: str):
        """
        Send a signed transaction to the network and return the hash of the transaction
        :param signed_tx: base64 encoded signed transaction, str.
        :param timeout: rpc request timeout
        :return:
        """
        try:
            return await self.json_rpc(
                "send_tx",
                {"signed_tx_base64": signed_tx, "wait_until": "INCLUDED"},
                broadcast=self.allow_broadcast,
            )
        except InvalidNonce:
            logger.warning("Invalid nonce during broadcast included transaction")
            return None

    async def wait_for_trx(self, trx_hash, receiver_id) -> TransactionResult:
        for _ in range(6):
            await asyncio.sleep(5)
            try:
                result = await self.get_tx(trx_hash, receiver_id)
            except InternalError:
                continue
            except Exception as e:
                logger.exception(e)
                continue
            if result:
                return result
        raise RPCTimeoutError("Transaction not found")

    async def send_tx_and_wait(
        self,
        signed_tx: str,
        trx_hash: Optional[str] = None,
        receiver_id: Optional[str] = None,
    ) -> TransactionResult:
        """
        Send a signed transaction to the network and wait for it to be included in a block
        :param signed_tx: base64 encoded signed transaction, str
        :param timeout: rpc request timeout
        :return:
        """
        try:
            res = await self.json_rpc(
                "broadcast_tx_commit",
                [signed_tx],
                broadcast=self.allow_broadcast,
            )
            return TransactionResult(**res)
        except RPCTimeoutError:
            if receiver_id and trx_hash:
                return await self.wait_for_trx(trx_hash, receiver_id)

    async def get_status(self):
        await self.check_available_rpcs()
        for rpc_addr in self._available_rpcs:
            try:
                data = {
                    "jsonrpc": "2.0",
                    "method": "status",
                    "params": {"finality": "final"},
                    "id": 1,
                }
                async with aiohttp.ClientSession() as session:
                    async with session.post(rpc_addr, json=data) as r:
                        if r.status == 200:
                            return json.loads(await r.text())["result"]
            except (
                ClientResponseError,
                ClientConnectorError,
                ServerDisconnectedError,
                ConnectionError,
            ) as e:
                logger.error(f"Rpc get status error: {e}")
            except Exception as e:
                logger.error(e)

        raise RpcNotAvailableError("RPC not available")

    async def get_validators(self):
        return await self.json_rpc("validators", [None])

    async def query(self, query_object):
        return await self.json_rpc("query", query_object)

    async def get_account(self, account_id, finality="optimistic"):
        return await self.json_rpc(
            "query",
            {
                "request_type": "view_account",
                "account_id": account_id,
                "finality": finality,
            },
        )

    async def get_access_key_list(self, account_id, finality="optimistic"):
        return await self.json_rpc(
            "query",
            {
                "request_type": "view_access_key_list",
                "account_id": account_id,
                "finality": finality,
            },
        )

    async def get_access_key(self, account_id, public_key, finality="optimistic"):
        """

        :param account_id:
        :param public_key:
        :param finality:
        :return: {'block_hash': '..', 'block_height': int, 'nonce': int, 'permission': 'FullAccess'}
        """
        return await self.json_rpc(
            "query",
            {
                "request_type": "view_access_key",
                "account_id": account_id,
                "public_key": public_key,
                "finality": finality,
            },
        )

    async def view_call(
        self,
        account_id,
        method_name,
        args,
        finality="optimistic",
        block_id: Optional[int] = None,
        threshold: Optional[int] = None,
    ):
        body = {
            "request_type": "call_function",
            "account_id": account_id,
            "method_name": method_name,
            "args_base64": base64.b64encode(args).decode("utf8"),
        }
        if block_id:
            body["block_id"] = block_id
        else:
            body["finality"] = finality
        return await self.json_rpc("query", body, threshold=threshold)

    async def get_block(self, block_id):
        return await self.json_rpc("block", [block_id])

    async def get_chunk(self, chunk_id):
        return await self.json_rpc("chunk", [chunk_id])

    async def get_tx(self, tx_hash, tx_recipient_id) -> TransactionResult:
        return TransactionResult(
            **await self.json_rpc("tx", [tx_hash, tx_recipient_id])
        )

    async def get_tx_status(self, tx_hash, tx_recipient_id) -> TransactionResult:
        return TransactionResult(
            **await self.json_rpc("EXPERIMENTAL_tx_status", [tx_hash, tx_recipient_id])
        )

    async def get_changes_in_block(self, changes_in_block_request):
        return await self.json_rpc(
            "EXPERIMENTAL_changes_in_block", changes_in_block_request
        )

    async def get_validators_ordered(self, block_hash):
        return await self.json_rpc("EXPERIMENTAL_validators_ordered", [block_hash])

    async def get_light_client_proof(
        self, outcome_type, tx_or_receipt_id, sender_or_receiver_id, light_client_head
    ):
        if outcome_type == "receipt":
            params = {
                "type": "receipt",
                "receipt_id": tx_or_receipt_id,
                "receiver_id": sender_or_receiver_id,
                "light_client_head": light_client_head,
            }
        else:
            params = {
                "type": "transaction",
                "transaction_hash": tx_or_receipt_id,
                "sender_id": sender_or_receiver_id,
                "light_client_head": light_client_head,
            }
        return await self.json_rpc("light_client_proof", params)

    async def get_next_light_client_block(self, last_block_hash):
        return await self.json_rpc("next_light_client_block", [last_block_hash])



NEAR = 1_000_000_000_000_000_000_000_000

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from py_near.account import Account


class DappClient:
    def __init__(self, account: "Account"):
        self._account = account




import base64
from typing import Union, List

import base58
import ed25519
from py_near_primitives import (
    CreateAccountAction,
    AddKeyAction,
    DeleteKeyAction,
    TransferAction,
    SignedDelegateAction,
    DelegateAction,
    DeployContractAction,
    FunctionCallAction,
    StakeAction,
    Transaction,
    AccessKey,
    AccessKeyPermissionFieldless,
    FunctionCallPermission,
)

from py_near.models import Action


def sign_and_serialize_transaction(
    account_id,
    private_key,
    receiver_id,
    nonce,
    actions: List[Action],
    block_hash: bytes,
) -> str:
    if isinstance(private_key, str):
        pk = base58.b58decode(private_key.replace("ed25519:", ""))
    else:
        pk = private_key
    private_key = ed25519.SigningKey(pk)

    transaction = Transaction(
        account_id,
        private_key.get_verifying_key().to_bytes(),
        nonce,
        receiver_id,
        block_hash,
        actions,
    )

    signed_trx = bytes(bytearray(transaction.to_vec(pk)))
    return base64.b64encode(signed_trx).decode("utf-8")


def calc_trx_hash(
    account_id,
    private_key,
    receiver_id,
    nonce,
    actions: List[Action],
    block_hash: bytes,
) -> str:
    if isinstance(private_key, str):
        pk = base58.b58decode(private_key.replace("ed25519:", ""))
    else:
        pk = private_key
    private_key = ed25519.SigningKey(pk)

    transaction = Transaction(
        account_id,
        private_key.get_verifying_key().to_bytes(),
        nonce,
        receiver_id,
        block_hash,
        actions,
    )

    signed_trx = bytes(bytearray(transaction.get_hash()))
    return base58.b58encode(signed_trx).decode("utf-8")


def create_create_account_action():
    return CreateAccountAction()


def create_full_access_key_action(pk: Union[bytes, str]):
    if isinstance(pk, str):
        pk = base58.b58decode(pk.replace("ed25519:", ""))
    return AddKeyAction(
        public_key=pk,
        access_key=AccessKey(0, AccessKeyPermissionFieldless.FullAccess),
    )


def create_function_call_access_key_action(
    pk, allowance: int, receiver_id: str, method_names: List[str]
):
    if isinstance(pk, str):
        pk = base58.b58decode(pk.replace("ed25519:", ""))
    return AddKeyAction(
        pk,
        AccessKey(0, FunctionCallPermission(receiver_id, method_names, allowance)),
    )


def create_delete_access_key_action(pk: Union[bytes, str]):
    if isinstance(pk, str):
        pk = base58.b58decode(pk.replace("ed25519:", ""))
    return DeleteKeyAction(pk)


def create_signed_delegate(action: DelegateAction, signature: bytes):
    return SignedDelegateAction(delegate_action=action, signature=signature)


def create_transfer_action(amount: int):
    return TransferAction(amount)


def create_staking_action(amount, pk: Union[bytes, str]):
    if isinstance(pk, str):
        pk = base58.b58decode(pk.replace("ed25519:", ""))
    return StakeAction(amount, pk)


def create_deploy_contract_action(code: bytes):
    return DeployContractAction(code)


def create_function_call_action(method_name: str, args, gas: int, deposit: int):
    return FunctionCallAction(method_name, args, gas, deposit)



import base64
import json
from dataclasses import dataclass, field
from enum import Enum
from json import JSONDecodeError
from typing import List, Any, Optional, Union

import base58
import py_near_primitives
from py_near_primitives.py_near_primitives import (
    DelegateAction,
    TransferAction,
    DeleteAccountAction,
    FunctionCallAction,
    DeployContractAction,
    CreateAccountAction,
    SignedDelegateAction,
    DeleteKeyAction,
    AddKeyAction,
    StakeAction,
)
from py_near_primitives.py_near_primitives import (
    FunctionCallPermission,
    AccessKeyPermissionFieldless,
)

from py_near.exceptions.exceptions import parse_error

Action = Union[
    DelegateAction,
    TransferAction,
    DeleteAccountAction,
    FunctionCallAction,
    DeployContractAction,
    CreateAccountAction,
    SignedDelegateAction,
    DeleteKeyAction,
    AddKeyAction,
    StakeAction,
]


class ReceiptOutcome:
    logs: List[str]
    metadata: dict
    receipt_ids: List[str]
    status: dict
    tokens_burnt: str
    executor_id: str
    gas_burnt: int

    def __init__(self, data):
        self.logs = data["outcome"]["logs"]
        self.metadata = data["outcome"]["metadata"]
        self.receipt_ids = data["outcome"]["receipt_ids"]
        self.status = data["outcome"]["status"]
        self.tokens_burnt = data["outcome"]["tokens_burnt"]
        self.gas_burnt = data["outcome"]["gas_burnt"]

    @property
    def error(self):
        if "Failure" in self.status:
            error_type, args = list(
                self.status["Failure"]["ActionError"]["kind"].items()
            )[0]
            return parse_error(error_type, args)


class ActionType(str, Enum):
    FUNCTION_CALL = "FunctionCall"
    TRANSFER = "Transfer"
    DELETE_ACCOUNT = "DeleteAccount"
    CREATE_ACCOUNT = "CreateAccount"
    ADD_KEY = "AddKey"
    STAKE = "Stake"
    DELETE_KEY = "DeleteKey"
    DEPLOY_CONTRACT = "DeployContract"
    DELEGATE = "Delegate"


class PublicKeyPermissionType(str, Enum):
    FULL_ACCESS = "FullAccess"
    FUNCTION_CALL = "FunctionCall"


@dataclass
class AccessKey:
    permission_type: PublicKeyPermissionType
    nonce: int
    allowance: Optional[str]
    receiver_id: Optional[str]
    method_names: Optional[List[str]]

    @classmethod
    def build(cls, data: dict) -> "AccessKey":
        if data["permission"] == PublicKeyPermissionType.FULL_ACCESS:
            return cls(
                nonce=data["nonce"], permission_type=PublicKeyPermissionType.FULL_ACCESS
            )

        permission_type, permission_data = list(data["permission"].items())[0]
        return cls(
            nonce=data["nonce"],
            permission_type=PublicKeyPermissionType.FUNCTION_CALL,
            **permission_data,
        )


@dataclass
class ReceiptDelegateAction:
    actions: List["ReceiptAction"]
    sender_id: str
    receiver_id: str
    public_key: str
    nonce: int
    max_block_height: int

    @classmethod
    def build(cls, data: dict) -> "ReceiptDelegateAction":
        actions = [ReceiptAction.build(action) for action in data["actions"]]
        del data["actions"]
        return cls(
            actions=actions,
            **data,
        )

    @property
    def near_delegate_action(self) -> DelegateAction:
        return DelegateAction(
            sender_id=self.sender_id,
            receiver_id=self.receiver_id,
            actions=[a.near_action for a in self.actions],
            nonce=self.nonce,
            max_block_height=self.max_block_height,
            public_key=base58.b58decode(self.public_key),
        )

    @property
    def nep461_hash(self) -> bytes:
        return bytes(bytearray(self.near_delegate_action.get_nep461_hash()))


@dataclass
class DelegateActionModel:
    actions: List[
        Union[
            DelegateAction,
            TransferAction,
            DeleteAccountAction,
            FunctionCallAction,
            DeployContractAction,
            CreateAccountAction,
            SignedDelegateAction,
            DeleteKeyAction,
            AddKeyAction,
            StakeAction,
        ]
    ]
    sender_id: str
    receiver_id: str
    public_key: str
    nonce: int
    max_block_height: int

    @property
    def near_delegate_action(self) -> DelegateAction:
        return DelegateAction(
            sender_id=self.sender_id,
            receiver_id=self.receiver_id,
            actions=self.actions,
            nonce=self.nonce,
            max_block_height=self.max_block_height,
            public_key=base58.b58decode(self.public_key),
        )

    @property
    def nep461_hash(self) -> bytes:
        return bytes(bytearray(self.near_delegate_action.get_nep461_hash()))

    @staticmethod
    def bytes_to_json(data: bytes) -> dict:
        return json.loads(DelegateAction.bytes_to_json(data))

    @classmethod
    def from_bytes(cls, data: bytes) -> "DelegateActionModel":
        data = cls.bytes_to_json(data)
        actions = [ReceiptAction.build(action) for action in data.pop("actions")]
        raw_actions = []
        for act in actions:
            if act.transactions_type == ActionType.FUNCTION_CALL:
                raw_actions.append(
                    FunctionCallAction(
                        act.method_name,
                        base64.b64decode(act.args_raw),
                        int(act.gas),
                        int(act.deposit),
                    )
                )
            elif act.transactions_type == ActionType.ADD_KEY:
                if (
                    act.access_key.permission_type
                    == PublicKeyPermissionType.FUNCTION_CALL
                ):
                    ak = py_near_primitives.py_near_primitives.AccessKey(
                        act.access_key.nonce,
                        FunctionCallPermission(
                            act.access_key.receiver_id,
                            act.access_key.method_names,
                            act.access_key.allowance,
                        ),
                    )
                else:
                    ak = AccessKey(
                        act.access_key.nonce, AccessKeyPermissionFieldless.FullAccess
                    )
                bpk = base58.b58decode(act.public_key.split(":")[1])
                raw_actions.append(AddKeyAction(bpk, ak))
            else:
                raise Exception("Undelegeted action type")
        return cls(**data, actions=raw_actions)


@dataclass
class ReceiptAction:
    transactions_type: ActionType
    # Transaction
    deposit: Optional[str] = None
    gas: Optional[str] = field(default=None)
    # DeployContract
    code: Optional[str] = field(default=None)
    # FunctionCall
    method_name: Optional[str] = field(default=None)
    args: Any = field(default=None)
    args_raw: Any = field(default=None)
    # DeleteAccount
    beneficiary_id: Optional[str] = field(default=None)
    # AddKey
    public_key: Optional[str] = field(default=None)
    access_key: Optional[AccessKey] = field(default=None)
    # Stake
    stake: Optional[int] = field(default=None)
    # Delegate
    signature: Optional[str] = field(default=None)
    delegate_action: Optional[ReceiptDelegateAction] = field(default=None)

    @classmethod
    def build(cls, data: dict) -> "ReceiptAction":
        if isinstance(data, str) and data == "CreateAccount":
            return cls(transactions_type=ActionType.CREATE_ACCOUNT)

        action_type, action_data = list(data.items())[0]
        access_key = None
        args = args_raw = None
        if action_type == ActionType.ADD_KEY:
            access_key = AccessKey.build(action_data["access_key"])
        elif action_type == ActionType.FUNCTION_CALL:
            args_raw = action_data["args"]
            try:
                args = base64.b64decode(action_data["args"])
                args = json.loads(args)
            except (UnicodeDecodeError, JSONDecodeError):
                args = None

        if action_type == ActionType.DELEGATE:
            delegate_action = ReceiptDelegateAction.build(
                action_data["delegate_action"]
            )
            return cls(
                transactions_type=action_type,
                signature=action_data["signature"],
                delegate_action=delegate_action,
            )

        action_data.pop("args", None)
        action_data.pop("access_key", None)
        return cls(
            transactions_type=action_type,
            access_key=access_key,
            args=args,
            args_raw=args_raw,
            **action_data,
        )


class TransactionData:
    hash: str
    public_key: str
    receiver_id: str
    signature: str
    signer_id: str
    nonce: int
    actions: List[ReceiptAction]

    def __init__(
        self,
        hash,
        public_key,
        receiver_id,
        signature,
        signer_id,
        nonce,
        actions,
        **kargs,
    ):
        self.actions = [ReceiptAction.build(a) for a in actions]
        self.nonce = nonce
        self.signer_id = signer_id
        self.public_key = public_key
        self.receiver_id = receiver_id
        self.signature = signature
        self.hash = hash

    @property
    def url(self):
        return f"https://nearblocks.io/ru/txns/{self.hash}"


class TransactionResult:
    receipt_outcome: List[ReceiptOutcome]
    receipts: List[dict] = []
    transaction_outcome: ReceiptOutcome
    status: dict
    transaction: TransactionData

    def __init__(
        self, receipts_outcome, transaction_outcome, transaction, status, **kargs
    ):
        self.status = status
        self.transaction = TransactionData(**transaction)
        self.transaction_outcome = ReceiptOutcome(transaction_outcome)
        if "receipts" in kargs:
            self.receipts = kargs["receipts"]

        self.receipt_outcome = []
        for ro in receipts_outcome:
            self.receipt_outcome.append(ReceiptOutcome(ro))

    @property
    def logs(self):
        logs = self.transaction_outcome.logs
        for ro in self.receipt_outcome:
            logs.extend(ro.logs)
        return logs


class ViewFunctionResult:
    block_hash: str
    block_height: str
    logs: List[str]
    result: Any

    def __init__(self, block_height, logs, result, block_hash=""):
        self.block_hash = block_hash
        self.block_height = block_height
        self.logs = logs
        self.result = result


@dataclass
class AccessKey:
    permission_type: PublicKeyPermissionType
    nonce: int
    allowance: Optional[str] = None
    receiver_id: Optional[str] = None
    method_names: Optional[List[str]] = None

    @classmethod
    def build(cls, data: dict) -> "AccessKey":
        if data["permission"] == PublicKeyPermissionType.FULL_ACCESS:
            return cls(
                nonce=data["nonce"], permission_type=PublicKeyPermissionType.FULL_ACCESS
            )

        permission_type, permission_data = list(data["permission"].items())[0]
        return cls(
            nonce=data["nonce"],
            permission_type=PublicKeyPermissionType.FUNCTION_CALL,
            **permission_data,
        )


@dataclass
class PublicKey:
    public_key: str
    access_key: AccessKey

    @classmethod
    def build(cls, data: dict) -> "PublicKey":
        return cls(
            data["public_key"],
            AccessKey.build(data["access_key"]),
        )


@dataclass
class AccountAccessKey:
    block_hash: str
    block_height: int
    nonce: int
    permission: Union[str, dict]



import json
from typing import Optional


class JsonProviderError(Exception):
    trx_hash: Optional[str] = None
    error_json: dict

    def __init__(self, *args, error_json=None, **kargs):
        self.error_json = error_json


class TransactionError(JsonProviderError):
    pass


class AccountError(JsonProviderError):
    pass


class BlockError(JsonProviderError):
    pass


class AccessKeyError(JsonProviderError):
    pass


class UnknownAccessKeyError(AccessKeyError):
    """
    The requested public_key has not been found while viewing since the
    public key has not been created or has been already deleted
    """

    pass


class UnknownBlockError(BlockError):
    """
    The requested block has not been produced yet or it has been
    garbage-collected (cleaned up to save space on the RPC node)
    """

    pass


class InternalError(TransactionError, AccountError, BlockError):
    """
    Something went wrong with the node itself or overloaded
    """

    pass


class NoSyncedYetError(BlockError):
    """
    The node is still syncing and the requested block is not in the database yet
    """

    pass


class InvalidAccount(AccountError):
    """
    The requested account_id is invalid
    """

    pass


class UnknownAccount(AccountError):
    """
    The requested account_id has not been found while viewing since the
    account has not been created or has been already deleted
    """

    pass


class NoContractCodeError(AccountError):
    """
    The account does not have any contract deployed on it
    """

    pass


class TooLargeContractStateError(AccountError):
    """
    The requested contract state is too large to be returned from this node
    (the default limit is 50kb of state size)
    """

    pass


class UnavailableShardError(AccountError):
    """
    The node was unable to find the requested data because it does not track the shard where data is present
    """

    pass


class NoSyncedBlocksError(AccountError):
    """
    The node is still syncing and the requested block is not in the database yet
    """

    pass


class InvalidTransactionError(TransactionError):
    """
    An error happened during transaction execution
    """

    pass


class TxExecutionError(InvalidTransactionError):
    def __init__(self, data={}, error_json=None, **kwargs):
        super().__init__(error_json=error_json)
        if isinstance(data, str):
            data = json.loads(data)
        data.update(kwargs)
        for key, value in data.items():
            setattr(self, key, value)


class InvalidTxError(TxExecutionError):
    pass


class ActionErrorKind(TxExecutionError):
    pass


class AccountAlreadyExists(ActionErrorKind):
    account_id: str


class AccountDoesNotExist(ActionErrorKind):
    account_id: str


class CreateAccountNotAllowed(ActionErrorKind):
    account_id: str
    predecessor_id: str


class ActorNoPermission(ActionErrorKind):
    account_id: str
    actor_id: str


class DeleteKeyDoesNotExist(ActionErrorKind):
    account_id: str
    public_key: str


class AddKeyAlreadyExists(ActionErrorKind):
    account_id: str
    public_key: str


class DeleteAccountStaking(ActionErrorKind):
    account_id: str


class DeleteAccountHasRent(ActionErrorKind):
    account_id: str
    balance: str


class RentUnpaid(ActionErrorKind):
    account_id: str
    amount: str


class TriesToUnstake(ActionErrorKind):
    account_id: str


class TriesToStake(ActionErrorKind):
    account_id: str
    stake: str
    locked: str
    balance: str


class FunctionCallError(ActionErrorKind):
    pass


class NewReceiptValidationError(ActionErrorKind):
    pass


_ACTION_ERROR_KINDS = {
    "AccountAlreadyExists": AccountAlreadyExists,
    "ActorNoPermission": ActorNoPermission,
    "CreateAccountNotAllowed": CreateAccountNotAllowed,
    "AccountDoesNotExist": AccountDoesNotExist,
    "DeleteKeyDoesNotExist": DeleteKeyDoesNotExist,
    "DeleteAccountStaking": DeleteAccountStaking,
    "AddKeyAlreadyExists": AddKeyAlreadyExists,
    "DeleteAccountHasRent": DeleteAccountHasRent,
    "RentUnpaid": RentUnpaid,
    "TriesToUnstake": TriesToUnstake,
    "TriesToStake": TriesToStake,
    "FunctionCallError": FunctionCallError,
    "ActionErrorKind": ActionErrorKind,
}


class ActionError(TxExecutionError):
    index: Optional[int]
    kind: ActionErrorKind

    def __init__(self, data, error_json=None):
        self.error_json = error_json
        if isinstance(data, str):
            data = json.loads(data)
        self.index = data.get("index", None)
        key, value = list(data["kind"].items())[0]
        self.kind = _ACTION_ERROR_KINDS[key](value)


class InvalidNonce(InvalidTxError):
    tx_nonce: int
    ak_nonce: int


class InvalidAccessKeyError(InvalidTxError):
    pass


class InvalidSignerId(InvalidTxError):
    signer_id: str


class SignerDoesNotExist(InvalidTxError):
    signer_id: str


class InvalidReceiverId(InvalidTxError):
    receiver_id: str


class NotEnoughBalance(InvalidTxError):
    signer_id: str
    balance: str
    cost: str


class LackBalanceForState(InvalidTxError):
    amount: str
    signer_id: str


class CostOverflow(InvalidTxError):
    pass


class InvalidChain(InvalidTxError):
    pass


class Expired(InvalidTxError):
    pass


class ActionsValidation(InvalidTxError):
    pass


class InvalidSignature(InvalidTxError):
    pass


class RPCTimeoutError(TransactionError):
    """
    Transaction was routed, but has not been recorded on chain in 10 seconds.
    """

    pass


# import inspect
# import sys
# clsmembers = inspect.getmembers(sys.modules[__name__], inspect.isclass)
# res = {}
# for c in clsmembers:
#     print(f"\"{c[0]}\": {c[0]},")


ERROR_CODE_TO_EXCEPTION = {
    "AccessKeyError": AccessKeyError,
    "AccountAlreadyExists": AccountAlreadyExists,
    "AccountDoesNotExist": AccountDoesNotExist,
    "AccountError": AccountError,
    "ActionError": ActionError,
    "ActionErrorKind": ActionErrorKind,
    "ActionsValidation": ActionsValidation,
    "ActorNoPermission": ActorNoPermission,
    "AddKeyAlreadyExists": AddKeyAlreadyExists,
    "BlockError": BlockError,
    "CostOverflow": CostOverflow,
    "CreateAccountNotAllowed": CreateAccountNotAllowed,
    "DeleteAccountHasRent": DeleteAccountHasRent,
    "DeleteAccountStaking": DeleteAccountStaking,
    "DeleteKeyDoesNotExist": DeleteKeyDoesNotExist,
    "Expired": Expired,
    "FunctionCallError": FunctionCallError,
    "InternalError": InternalError,
    "InvalidAccessKeyError": InvalidAccessKeyError,
    "InvalidAccount": InvalidAccount,
    "InvalidChain": InvalidChain,
    "InvalidNonce": InvalidNonce,
    "InvalidReceiverId": InvalidReceiverId,
    "InvalidSignature": InvalidSignature,
    "InvalidSignerId": InvalidSignerId,
    "InvalidTransactionError": InvalidTransactionError,
    "InvalidTxError": InvalidTxError,
    "JsonProviderError": JsonProviderError,
    "NewReceiptValidationError": NewReceiptValidationError,
    "NoContractCodeError": NoContractCodeError,
    "NoSyncedBlocksError": NoSyncedBlocksError,
    "NoSyncedYetError": NoSyncedYetError,
    "NotEnoughBalance": NotEnoughBalance,
    "LackBalanceForState": LackBalanceForState,
    "RentUnpaid": RentUnpaid,
    "RpcTimeoutError": RPCTimeoutError,
    "SignerDoesNotExist": SignerDoesNotExist,
    "TooLargeContractStateError": TooLargeContractStateError,
    "TransactionError": TransactionError,
    "TriesToStake": TriesToStake,
    "TriesToUnstake": TriesToUnstake,
    "TxExecutionError": TxExecutionError,
    "UnavailableShardError": UnavailableShardError,
    "UnknownAccessKeyError": UnknownAccessKeyError,
    "UnknownAccount": UnknownAccount,
    "UnknownBlockError": UnknownBlockError,
}
